#include <source_location>
#include <jlog/ansi_escape_codes.hpp>
#include <jlog/jlog.hpp>
#include <string>
#include <chrono>
#include <fstream>
#include <iostream>
#include <windows.h>

namespace jlog {
    std::string get_timestamp() {
        using namespace std::chrono;
        auto const timestamp = current_zone()->to_local(system_clock::now());
        auto dp = floor<days>(timestamp);
        year_month_day ymd{floor<days>(timestamp)};
        hh_mm_ss time{floor<milliseconds>(timestamp - dp)};
        auto y = ymd.year();
        auto m = ymd.month();
        auto d = ymd.day();
        auto h = time.hours();
        auto M = time.minutes();
        auto s = time.seconds();
        auto ms = time.subseconds();
        return std::format("{}-{}-{} {}:{}:{}.{}", y, m, d, h.count(), M.count(), s.count(), ms.count());
    }

    void log_to_console(const std::string& message) {
        std::cout << message;
    }

    void log_to_file(const std::string& message) {
        std::ofstream latest_log("latest.log", std::ios_base::app);
        latest_log << message;
        latest_log.close();
    }

    void log(std::vector<token> tokens) {
        for (const token& t : tokens) {
            if (!t.delimiter.empty()) {
                ansi_escape_codes::SetConsoleTextColor(t.colorCode);
                log_to_console(std::string(1, t.delimiter[0]) + t.content + std::string(1, t.delimiter[1]) + " ");
                log_to_file(std::string(1, t.delimiter[0]) + t.content + std::string(1, t.delimiter[1]) + " ");
            } else {
                ansi_escape_codes::SetConsoleTextColor(t.colorCode);
                log_to_console(t.content + " ");
                log_to_file(t.content + " ");
            }
        }
        ansi_escape_codes::SetConsoleTextColor(ansi_escape_codes::FG_DEFAULT);
        log_to_console("\n");
        log_to_file("\n");
    }

    void direct(const std::string& message) {
        log({{.content = message, .delimiter = ""}});
    }

    void info(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void sinfo(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void usinfo(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void verbose(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void sverbose(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void usverbose(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void debug(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void sdebug(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void usdebug(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void warning(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void swarning(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void uswarning(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void error(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void serror(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void userror(const std::string& message) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void fatal(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        log({timestamp, severity, content});
    }

    void sfatal(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void usfatal(const std::string& message) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void info_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void sinfo_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void usinfo_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_WHITE, .content = "INFO"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void verbose_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void sverbose_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void usverbose_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_CYAN, .content = "VERBOSE"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void debug_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void sdebug_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void usdebug_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_GREEN, .content = "DEBUG"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void warning_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void swarning_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void uswarning_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_YELLOW, .content = "WARNING"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void error_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void serror_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void userror_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_RED, .content = "ERROR"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }

    void fatal_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({timestamp, trace, filedata, severity, content});
    }

    void sfatal_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto timestamp = token{.content = get_timestamp()};
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        log({trace, filedata, severity, content});
    }

    void usfatal_spec(const std::string& message, const std::string& func, const std::string& file, int line) {
        auto severity = token{.colorCode = ansi_escape_codes::FG_BRIGHT_RED, .content = "FATAL"};
        auto content = token{.content = message, .delimiter = ""};
        log({severity, content});
    }
}